"""
Tail the specified log files.
"""

from fabric.api import env, execute, parallel, run
from pkg_resources import parse_version

from .common import (
    add_config,
    add_environment,
    add_name,
    add_override_name,
    add_pattern,
    add_pool_size,
)
from ..util import (
    activate_env,
    get_env_config,
    get_logfiles_cmd,
    get_topology_definition,
    get_nimbus_client,
    nimbus_storm_version,
    ssh_tunnel,
)


@parallel
def _tail_logs(topology_name, pattern, follow, num_lines, is_old_storm):
    """
    Actual task to run tail on all servers in parallel.
    """
    ls_cmd = get_logfiles_cmd(
        topology_name=topology_name, pattern=pattern, is_old_storm=is_old_storm
    )
    tail_pipe = f" | xargs tail -n {num_lines}"
    if follow:
        tail_pipe += " -f"
    run(ls_cmd + tail_pipe)


def tail_topology(
    topology_name=None,
    env_name=None,
    pattern=None,
    follow=False,
    num_lines=10,
    override_name=None,
    config_file=None,
):
    """Follow (tail -f) the log files on remote Storm workers.

    Will use the `log_path` and `workers` properties from config.json.
    """
    if override_name is not None:
        topology_name = override_name
    else:
        topology_name = get_topology_definition(topology_name, config_file=config_file)[
            0
        ]
    env_name, env_config = get_env_config(env_name, config_file=config_file)
    with ssh_tunnel(env_config) as (host, port):
        nimbus_client = get_nimbus_client(env_config, host=host, port=port)
        is_old_storm = nimbus_storm_version(nimbus_client) < parse_version("1.0")
    activate_env(env_name)
    execute(
        _tail_logs,
        topology_name,
        pattern,
        follow,
        num_lines,
        is_old_storm,
        hosts=env.storm_workers,
    )


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser("tail", description=__doc__, help=main.__doc__)
    subparser.set_defaults(func=main)
    add_config(subparser)
    add_environment(subparser)
    subparser.add_argument(
        "-f",
        "--follow",
        action="store_true",
        help="Keep files open and append output as they "
        'grow.  This is the same as "-f" parameter for '
        "the tail command that will be executed on the "
        "Storm workers.",
    )
    subparser.add_argument(
        "-l",
        "--num_lines",
        default=10,
        help="tail outputs the last NUM_LINES lines of the "
        "logs. (default: %(default)s)",
    )
    add_name(subparser)
    add_override_name(subparser)
    add_pool_size(subparser)
    add_pattern(subparser)


def main(args):
    """ Tail logs for specified Storm topology. """
    env.pool_size = args.pool_size
    tail_topology(
        topology_name=args.name,
        env_name=args.environment,
        pattern=args.pattern,
        follow=args.follow,
        num_lines=args.num_lines,
        override_name=args.override_name,
        config_file=args.config,
    )



"""
Run a local Storm topology.

Note: If you have "org.apache.storm" in your uberjar-exclusions list in your
project.clj, this will fail. Temporarily remove it to use `sparse run`. You will
also need to add [org.apache.storm/flux-core "1.0.1"] to dependencies.
"""

from argparse import RawDescriptionHelpFormatter
from tempfile import NamedTemporaryFile

from fabric.api import local, show
from ruamel import yaml

from ..util import (
    get_config,
    get_env_config,
    get_topology_definition,
    get_topology_from_file,
    local_storm_version,
    set_topology_serializer,
    storm_lib_version,
)
from .common import (
    add_ackers,
    add_config,
    add_debug,
    add_environment,
    add_name,
    add_options,
    add_workers,
    resolve_options,
)
from .jar import jar_for_deploy


def run_local_topology(
    name=None, env_name=None, time=0, options=None, config_file=None
):
    """Run a topology locally using Flux and `storm jar`."""
    name, topology_file = get_topology_definition(name, config_file=config_file)
    config = get_config(config_file=config_file)
    env_name, env_config = get_env_config(env_name, config_file=config_file)
    topology_class = get_topology_from_file(topology_file)

    set_topology_serializer(env_config, config, topology_class)

    storm_options = resolve_options(
        options, env_config, topology_class, name, local_only=True
    )
    if storm_options["topology.acker.executors"] != 0:
        storm_options["topology.acker.executors"] = 1
    storm_options["topology.workers"] = 1

    # Set parallelism based on env_name if necessary
    for spec in topology_class.specs:
        if isinstance(spec.par, dict):
            spec.par = spec.par.get(env_name)

    # Check Storm version is the same
    local_version = local_storm_version()
    project_version = storm_lib_version()
    if local_version != project_version:
        raise ValueError(
            "Local Storm version, {}, is not the same as the "
            "version in your project.clj, {}. The versions must "
            "match.".format(local_version, project_version)
        )

    # Prepare a JAR that has Storm dependencies packaged
    topology_jar = jar_for_deploy(simple_jar=False)

    if time <= 0:
        time = 9223372036854775807  # Max long value in Java

    # Write YAML file
    with show("output"):
        with NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as yaml_file:
            topology_flux_dict = topology_class.to_flux_dict(name)
            topology_flux_dict["config"] = storm_options
            if yaml.version_info < (0, 15):
                yaml.safe_dump(topology_flux_dict, yaml_file, default_flow_style=False)
            else:
                yml = yaml.YAML(typ="safe", pure=True)
                yml.default_flow_style = False
                yml.dump(topology_flux_dict, yaml_file)
            cmd = (
                "storm jar {jar} org.apache.storm.flux.Flux --local --no-splash "
                "--sleep {time} {yaml}".format(
                    jar=topology_jar, time=time, yaml=yaml_file.name
                )
            )
            local(cmd)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser(
        "run",
        description=__doc__,
        help=main.__doc__,
        formatter_class=RawDescriptionHelpFormatter,
    )
    subparser.set_defaults(func=main)
    add_ackers(subparser)
    add_config(subparser)
    add_debug(subparser)
    add_environment(subparser)
    add_name(subparser)
    add_options(subparser)
    subparser.add_argument(
        "-t",
        "--time",
        default=0,
        type=int,
        help="Time (in seconds) to keep local cluster "
        "running. If time <= 0, run indefinitely. "
        "(default: %(default)s)",
    )
    add_workers(subparser)


def main(args):
    """Run the local topology with the given arguments"""
    run_local_topology(
        name=args.name,
        time=args.time,
        options=args.options,
        env_name=args.environment,
        config_file=args.config,
    )



"""
Display uptime for workers in running Storm topologies.
"""

from pkg_resources import parse_version

from .common import add_config, add_environment
from ..util import get_ui_json, get_ui_jsons, print_stats_table, storm_lib_version


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser(
        "worker_uptime", description=__doc__, help=main.__doc__
    )
    subparser.set_defaults(func=main)
    add_config(subparser)
    add_environment(subparser)


def display_worker_uptime(env_name, config_file=None):
    topology_summary_path = "/api/v1/topology/summary"
    topology_detail_path = "/api/v1/topology/{topology}"
    component_path = "/api/v1/topology/{topology}/component/{component}"
    topo_summary_json = get_ui_json(
        env_name, topology_summary_path, config_file=config_file
    )
    topology_ids = [x["id"] for x in topo_summary_json["topologies"]]
    topology_components = dict()
    worker_stats = []
    topology_detail_jsons = get_ui_jsons(
        env_name,
        (topology_detail_path.format(topology=topology) for topology in topology_ids),
        config_file=config_file,
    )

    for topology in topology_ids:
        topology_detail_json = topology_detail_jsons[
            topology_detail_path.format(topology=topology)
        ]
        spouts = [x["spoutId"] for x in topology_detail_json["spouts"]]
        bolts = [x["boltId"] for x in topology_detail_json["bolts"]]
        topology_components[topology] = spouts + bolts

    comp_details = get_ui_jsons(
        env_name,
        (
            component_path.format(topology=topology, component=comp)
            for topology, comp_list in topology_components.items()
            for comp in comp_list
        ),
        config_file=config_file,
    )

    for comp_detail in comp_details.values():
        worker_stats += [
            (worker["host"], worker["id"], worker["uptime"], worker["workerLogLink"])
            for worker in comp_detail["executorStats"]
        ]
    worker_stats = sorted(set(worker_stats))

    print_stats_table(
        "Worker Stats",
        worker_stats,
        ["Host", "Worker ID", "Uptime", "Log URL"],
        custom_alignment={"Uptime": "r"},
    )


def main(args):
    """ Display uptime for Storm workers. """
    storm_version = storm_lib_version()
    if storm_version >= parse_version("0.9.2-incubating"):
        display_worker_uptime(args.environment, config_file=args.config)
    else:
        print(f"ERROR: Storm {storm_version} does not support this command.")



"""
Create or update a virtualenv on the Storm workers.  This will be done
automatically upon submit, but this command is provided to help with testing
and debugging.
"""

import os

from fabric.api import env, execute, parallel, put, puts, run, show
from fabric.contrib.files import exists

from .common import (
    add_config,
    add_environment,
    add_name,
    add_options,
    add_override_name,
    add_overwrite_virtualenv,
    add_pool_size,
    add_requirements,
    add_user,
    resolve_options,
    warn_about_deprecated_user,
)
from ..util import (
    activate_env,
    die,
    get_config,
    get_env_config,
    get_topology_definition,
    get_topology_from_file,
    run_cmd,
)


@parallel
def _create_or_update_virtualenv(
    virtualenv_root,
    virtualenv_name,
    requirements_paths,
    virtualenv_flags=None,
    overwrite_virtualenv=False,
    user=None,
):
    with show("output"):
        virtualenv_path = "/".join((virtualenv_root, virtualenv_name))
        if overwrite_virtualenv:
            puts(f"Removing virtualenv if it exists in {virtualenv_root}")
            cmd = f"rm -rf {virtualenv_path}"
            run_cmd(cmd, user, warn_only=True)
        if not exists(virtualenv_path):
            if virtualenv_flags is None:
                virtualenv_flags = ""
            puts(f"virtualenv not found in {virtualenv_root}, creating one.")
            cmd = f"virtualenv --never-download {virtualenv_path} {virtualenv_flags}"
            run_cmd(cmd, user)

        if isinstance(requirements_paths, str):
            requirements_paths = [requirements_paths]
        temp_req_paths = []
        for requirements_path in requirements_paths:
            puts(f"Uploading {requirements_path} to temporary file.")
            temp_req = run("mktemp /tmp/streamparse_requirements-XXXXXXXXX.txt")
            temp_req_paths.append(temp_req)
            put(requirements_path, temp_req, mode="0666")

        puts(f"Updating virtualenv: {virtualenv_name}")
        pip_path = "/".join((virtualenv_path, "bin", "pip"))
        # Make sure we're using latest pip so options work as expected
        run_cmd(f"{pip_path} install --upgrade 'pip>=9.0,!=19.0' setuptools", user)
        run_cmd(
            (
                "{} install -r {} --exists-action w --upgrade "
                "--upgrade-strategy only-if-needed --progress-bar off"
            ).format(pip_path, " -r ".join(temp_req_paths)),
            user,
        )

        run(f"rm -f {' '.join(temp_req_paths)}")


def create_or_update_virtualenvs(
    env_name,
    topology_name,
    options,
    virtualenv_name=None,
    requirements_paths=None,
    config_file=None,
    overwrite_virtualenv=False,
    user=None,
):
    """Create or update virtualenvs on remote servers.

    Assumes that virtualenv is on the path of the remote server(s).

    :param env_name: the name of the environment in config.json.
    :param topology_name: the name of the topology (and virtualenv).
    :param virtualenv_name: the name that we should use for the virtualenv, even
                          though the topology file has a different name.
    :param requirements_paths: a list of paths to requirements files to use to
                               create virtualenv
    :param overwrite_virtualenv: Force the creation of a fresh virtualenv, even
                                 if it already exists.
    :param user: Who to delete virtualenvs as when using overwrite_virtualenv
    """
    warn_about_deprecated_user(user, "create_or_update_virtualenvs")
    config = get_config()
    topology_name, topology_file = get_topology_definition(
        topology_name, config_file=config_file
    )
    topology_class = get_topology_from_file(topology_file)
    env_name, env_config = get_env_config(env_name, config_file=config_file)
    if virtualenv_name is None:
        virtualenv_name = topology_name

    config["virtualenv_specs"] = config["virtualenv_specs"].rstrip("/")

    if requirements_paths is None:
        requirements_paths = [
            os.path.join(config["virtualenv_specs"], f"{topology_name}.txt")
        ]

    # Check to ensure streamparse is in at least one requirements file
    found_streamparse = False
    for requirements_path in requirements_paths:
        with open(requirements_path) as fp:
            for line in fp:
                if "streamparse" in line:
                    found_streamparse = True
                    break

    if not found_streamparse:
        die(
            "Could not find streamparse in one of your requirements files.  "
            "Checked {}.  streamparse is required for all topologies.".format(
                requirements_paths
            )
        )

    # Setup the fabric env dictionary
    storm_options = resolve_options(options, env_config, topology_class, topology_name)
    activate_env(env_name, storm_options, config_file=config_file)

    # Actually create or update virtualenv on worker nodes
    execute(
        _create_or_update_virtualenv,
        env.virtualenv_root,
        virtualenv_name,
        requirements_paths,
        virtualenv_flags=storm_options.get("virtualenv_flags"),
        hosts=env.storm_workers,
        overwrite_virtualenv=overwrite_virtualenv,
        user=user or storm_options["sudo_user"],
    )


def subparser_hook(subparsers):
    """Hook to add subparser for this command."""
    subparser = subparsers.add_parser(
        "update_virtualenv", description=__doc__, help=main.__doc__
    )
    subparser.set_defaults(func=main)
    add_config(subparser)
    add_environment(subparser)
    add_overwrite_virtualenv(subparser)
    add_name(subparser)
    add_options(subparser)
    add_override_name(subparser)
    add_pool_size(subparser)
    add_requirements(subparser)
    add_user(subparser)


def main(args):
    """Create or update a virtualenv on Storm workers."""
    env.pool_size = args.pool_size
    create_or_update_virtualenvs(
        args.environment,
        args.name,
        args.options,
        virtualenv_name=args.override_name,
        requirements_paths=args.requirements,
        config_file=args.config,
        overwrite_virtualenv=args.overwrite_virtualenv,
    )



"""
This module provides the base sparse command and a load hook for dynamically
adding other subcommands.  The "load_suparsers" function searches for modules
in the streamparse/bin directory that have a "subparser_hook" method. The
"subparser_hook" accepts a the sparse subparsers object and adds it's
subparser as needed.
"""

import argparse
import importlib
import os
import pkgutil
import sys

from ..util import die
from ..version import __version__


def load_subparsers(subparsers):
    """
    searches modules in streamparse/bin for a 'subparser_hook' method and calls
    the 'subparser_hook' method on the sparse subparsers object.
    """
    for _, mod_name, is_pkg in pkgutil.iter_modules([os.path.dirname(__file__)]):
        if not is_pkg and mod_name not in sys.modules:
            module = importlib.import_module(f"streamparse.cli.{mod_name}")
            # check for the subparser hook
            if hasattr(module, "subparser_hook"):
                module.subparser_hook(subparsers)


def main():
    """main entry point for sparse"""
    parser = argparse.ArgumentParser(
        description="Utilities for managing Storm" "/streamparse topologies.",
        epilog="sparse provides a front-end to "
        "streamparse, a framework for "
        "creating Python projects for "
        "running, debugging, and "
        "submitting computation topologies "
        "against real-time streams, using "
        "Apache Storm. It requires java and"
        " lein (Clojure build tool) to be "
        "on your $PATH, and uses lein and "
        "Clojure under the hood for JVM/"
        "Thrift interop.",
    )
    subparsers = parser.add_subparsers(title="sub-commands")
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    load_subparsers(subparsers)

    def _help_command(args):
        """Print help information about other commands.

        Does the same thing as adding --help flag to sub-command calls.
        """
        subparsers.choices[args.sub_command].print_help()
        sys.exit(1)

    help_parser = subparsers.add_parser(
        "help",
        description=_help_command.__doc__,
        help=_help_command.__doc__.splitlines()[0],
    )
    help_parser.add_argument(
        "sub_command",
        help="The command to provide help for.",
        choices=sorted(subparsers.choices.keys()),
    )
    help_parser.set_defaults(func=_help_command)
    args = parser.parse_args()

    if os.getuid() == 0 and not os.getenv("LEIN_ROOT"):
        die(
            "Because streamparse relies on Leiningen, you cannot run "
            "streamparse as root without the LEIN_ROOT environment variable "
            "set. Otherwise, Leiningen would hang indefinitely under-the-hood "
            "waiting for user input."
        )

    # http://grokbase.com/t/python/python-bugs-list/12arsq9ayf/issue16308-undocumented-behaviour-change-in-argparse-from-3-2-3-to-3-3-0
    if hasattr(args, "func"):
        args.func(args)
    # python3.3+ argparse changes
    else:
        parser.print_help()
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
Display stats about running Storm topologies.
"""

import sys
from itertools import chain

from pkg_resources import parse_version

from ..util import (
    get_env_config,
    get_ui_json,
    get_ui_jsons,
    print_stats_table,
    storm_lib_version,
)
from .common import add_config, add_environment, add_name


def display_stats(
    env_name,
    topology_name=None,
    component_name=None,
    all_components=None,
    config_file=None,
):
    env_name = env_name
    if topology_name and all_components:
        _print_all_components(env_name, topology_name, config_file=config_file)
    elif topology_name and component_name:
        _print_component_status(
            env_name, topology_name, component_name, config_file=config_file
        )
    elif topology_name:
        _print_topology_status(env_name, topology_name, config_file=config_file)
    else:
        _print_cluster_status(env_name, config_file=config_file)


def _print_cluster_status(env_name, config_file=None):
    jsons = get_ui_jsons(
        env_name,
        [
            "/api/v1/cluster/summary",
            "/api/v1/topology/summary",
            "/api/v1/supervisor/summary",
        ],
        config_file=config_file,
    )
    # Print Cluster Summary
    ui_cluster_summary = jsons["/api/v1/cluster/summary"]
    columns = [
        "stormVersion",
        "nimbusUptime",
        "supervisors",
        "slotsTotal",
        "slotsUsed",
        "slotsFree",
        "executorsTotal",
        "tasksTotal",
    ]
    print_stats_table("Cluster summary", ui_cluster_summary, columns, "r")
    # Print Topologies Summary
    ui_topologies_summary = jsons["/api/v1/topology/summary"]
    columns = [
        "name",
        "id",
        "status",
        "uptime",
        "workersTotal",
        "executorsTotal",
        "tasksTotal",
    ]
    print_stats_table(
        "Topology summary", ui_topologies_summary["topologies"], columns, "r"
    )
    # Print Supervisor Summary
    ui_supervisor_summary = jsons["/api/v1/supervisor/summary"]
    columns = ["id", "host", "uptime", "slotsTotal", "slotsUsed"]
    print_stats_table(
        "Supervisor summary",
        ui_supervisor_summary["supervisors"],
        columns,
        "r",
        {"host": "l", "uptime": "l"},
    )


def _get_topology_ui_detail(env_name, topology_name, config_file=None):
    env_name = get_env_config(env_name, config_file=config_file)[0]
    topology_id = _get_topology_id(env_name, topology_name)
    detail_url = f"/api/v1/topology/{topology_id}"
    detail = get_ui_json(env_name, detail_url, config_file=config_file)
    return detail


def _print_topology_status(env_name, topology_name, config_file=None):
    ui_detail = _get_topology_ui_detail(
        env_name, topology_name, config_file=config_file
    )
    # Print topology summary
    columns = [
        "name",
        "id",
        "status",
        "uptime",
        "workersTotal",
        "executorsTotal",
        "tasksTotal",
    ]
    print_stats_table("Topology summary", ui_detail, columns, "r")
    # Print topology stats
    columns = [
        "windowPretty",
        "emitted",
        "transferred",
        "completeLatency",
        "acked",
        "failed",
    ]
    print_stats_table("Topology stats", ui_detail["topologyStats"], columns, "r")
    # Print spouts
    if ui_detail.get("spouts"):
        columns = [
            "spoutId",
            "emitted",
            "transferred",
            "completeLatency",
            "acked",
            "failed",
        ]
        print_stats_table(
            "Spouts (All time)", ui_detail["spouts"], columns, "r", {"spoutId": "l"}
        )

    columns = [
        "boltId",
        "executors",
        "tasks",
        "emitted",
        "transferred",
        "capacity",
        "executeLatency",
        "executed",
        "processLatency",
        "acked",
        "failed",
        "lastError",
    ]
    print_stats_table(
        "Bolt (All time)", ui_detail["bolts"], columns, "r", {"boltId": "l"}
    )


def _get_component_ui_detail(
    env_name, topology_name, component_names, config_file=None
):
    if isinstance(component_names, str):
        component_names = [component_names]
    env_name = get_env_config(env_name, config_file=config_file)[0]
    topology_id = _get_topology_id(env_name, topology_name, config_file=config_file)
    base_url = "/api/v1/topology/%s/component/%s"
    detail_urls = [base_url % (topology_id, name) for name in component_names]
    detail = get_ui_jsons(env_name, detail_urls, config_file=config_file)
    if len(detail) == 1:
        return list(detail.values())[0]
    else:
        return detail


def _print_all_components(env_name, topology_name, config_file=None):
    topology_ui_detail = _get_topology_ui_detail(env_name, topology_name)
    spouts = map(lambda spout: (spout["spoutId"], topology_ui_detail.get("spouts", {})))
    bolts = map(lambda spout: (spout["boltId"], topology_ui_detail.get("bolts", {})))
    ui_details = _get_component_ui_detail(
        env_name, topology_name, chain(spouts, bolts), config_file=config_file
    )
    names_and_keys = zip(
        map(lambda ui_detail: (ui_detail["name"], ui_details.values())),
        ui_details.keys(),
    )
    for component_name, key in names_and_keys:
        _print_component_status(
            env_name,
            topology_name,
            component_name,
            ui_details[key],
            config_file=config_file,
        )


def _print_component_status(
    env_name, topology_name, component_name, ui_detail=None, config_file=None
):
    if not ui_detail:
        ui_detail = _get_component_ui_detail(
            env_name, topology_name, component_name, config_file=config_file
        )
    _print_component_summary(ui_detail)
    if ui_detail.get("componentType") == "spout":
        _print_spout_stats(ui_detail)
        _print_spout_output_stats(ui_detail)
        _print_spout_executors(ui_detail)
    elif ui_detail.get("componentType") == "bolt":
        _print_bolt_stats(ui_detail)
        _print_input_stats(ui_detail)
        _print_bolt_output_stats(ui_detail)


def _print_component_summary(ui_detail):
    columns = ["id", "name", "executors", "tasks"]
    print_stats_table("Component summary", ui_detail, columns, "r")


def _print_bolt_stats(ui_detail):
    columns = [
        "windowPretty",
        "emitted",
        "transferred",
        "executeLatency",
        "executed",
        "processLatency",
        "acked",
        "failed",
    ]

    print_stats_table(
        "Bolt stats", ui_detail["boltStats"], columns, "r", {"windowPretty": "l"}
    )


def _print_input_stats(ui_detail):
    columns = [
        "component",
        "stream",
        "executeLatency",
        "processLatency",
        "executed",
        "acked",
        "failed",
    ]
    if ui_detail["inputStats"]:
        print_stats_table(
            "Input stats (All time)",
            ui_detail["inputStats"],
            columns,
            "r",
            {"component": "l"},
        )


def _print_bolt_output_stats(ui_detail):
    if ui_detail["outputStats"]:
        columns = ["stream", "emitted", "transferred"]
        print_stats_table(
            "Output stats (All time)",
            ui_detail["outputStats"],
            columns,
            "r",
            {"stream": "l"},
        )


def _print_spout_stats(ui_detail):
    columns = [
        "windowPretty",
        "emitted",
        "transferred",
        "completeLatency",
        "acked",
        "failed",
    ]
    data = ui_detail["spoutSummary"][-1].copy()
    print_stats_table("Spout stats", data, columns, "r", {"windowPretty": "l"})


def _print_spout_output_stats(ui_detail):
    columns = ["stream", "emitted", "transferred", "completeLatency", "acked", "failed"]
    print_stats_table(
        "Output stats (All time)",
        ui_detail["outputStats"],
        columns,
        "r",
        {"stream": "l"},
    )


def _print_spout_executors(ui_detail):
    columns = [
        "id",
        "uptime",
        "host",
        "port",
        "emitted",
        "transferred",
        "completeLatency",
        "acked",
        "failed",
    ]
    print_stats_table(
        "Executors (All time)", ui_detail["executorStats"], columns, "r", {"host": "l"}
    )


def _get_topology_id(env_name, topology_name, config_file=None):
    """Get toplogy ID from summary json provided by UI api"""
    summary_url = "/api/v1/topology/summary"
    topology_summary = get_ui_json(env_name, summary_url, config_file=config_file)
    for topology in topology_summary["topologies"]:
        if topology_name == topology["name"]:
            return topology["id"]


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser("stats", description=__doc__, help=main.__doc__)
    subparser.set_defaults(func=main)
    subparser.add_argument("--all", action="store_true", help="All available stats.")
    subparser.add_argument(
        "-c",
        "--component",
        help="Topology component (bolt/spout) name as "
        "specified in Clojure topology specification",
    )
    add_config(subparser)
    add_environment(subparser)
    add_name(subparser)


def main(args):
    """ Display stats about running Storm topologies. """
    storm_version = storm_lib_version()
    if storm_version >= parse_version("0.9.2-incubating"):
        display_stats(
            args.environment,
            topology_name=args.name,
            component_name=args.component,
            all_components=args.all,
            config_file=args.config,
        )
    else:
        print(f"ERROR: Storm {storm_version} does not support this command.")
        sys.stdout.flush()



"""
Remove all logs from Storm workers for the specified Storm topology.
"""

from fabric.api import env, execute, parallel

from .common import (
    add_config,
    add_environment,
    add_name,
    add_override_name,
    add_pattern,
    add_pool_size,
    add_user,
    resolve_options,
    warn_about_deprecated_user,
)
from ..util import (
    activate_env,
    get_env_config,
    get_topology_definition,
    get_topology_from_file,
    get_logfiles_cmd,
    run_cmd,
)


@parallel
def _remove_logs(
    topology_name, pattern, remove_worker_logs, user, remove_all_artifacts
):
    """
    Actual task to remove logs on all servers in parallel.
    """
    ls_cmd = get_logfiles_cmd(
        topology_name=topology_name,
        pattern=pattern,
        include_worker_logs=remove_worker_logs,
        include_all_artifacts=remove_all_artifacts,
    )
    rm_pipe = " | xargs rm -f"
    run_cmd(ls_cmd + rm_pipe, user, warn_only=True)


def remove_logs(
    topology_name=None,
    env_name=None,
    pattern=None,
    remove_worker_logs=False,
    user=None,
    override_name=None,
    remove_all_artifacts=False,
    options=None,
    config_file=None,
):
    """Remove all Python logs on Storm workers in the log.path directory."""
    warn_about_deprecated_user(user, "remove_logs")
    topology_name, topology_file = get_topology_definition(
        override_name or topology_name, config_file=config_file
    )
    topology_class = get_topology_from_file(topology_file)
    env_name, env_config = get_env_config(env_name, config_file=config_file)
    storm_options = resolve_options(options, env_config, topology_class, topology_name)
    activate_env(env_name, storm_options, config_file=config_file)
    execute(
        _remove_logs,
        topology_name,
        pattern,
        remove_worker_logs,
        # TODO: Remove "user" in next major version
        user or storm_options["sudo_user"],
        remove_all_artifacts,
        hosts=env.storm_workers,
    )


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser(
        "remove_logs", description=__doc__, help=main.__doc__
    )
    subparser.set_defaults(func=main)
    subparser.add_argument(
        "-A",
        "--remove_all_artifacts",
        help="Remove not only topology-specific logs, but "
        "also any other files for the topology in its "
        "workers-artifacts subdirectories.",
        action="store_true",
    )
    add_config(subparser)
    add_environment(subparser)
    add_name(subparser)
    add_override_name(subparser)
    add_pattern(subparser)
    add_pool_size(subparser)
    add_user(subparser, allow_short=True)
    subparser.add_argument(
        "-w",
        "--remove_worker_logs",
        help="Remove not only topology-specific logs, but "
        "also worker logs that may be shared between "
        "topologies.",
        action="store_true",
    )


def main(args):
    """ Remove logs from Storm workers. """
    env.pool_size = args.pool_size
    remove_logs(
        topology_name=args.name,
        env_name=args.environment,
        pattern=args.pattern,
        remove_worker_logs=args.remove_worker_logs,
        options=args.options,
        override_name=args.override_name,
        remove_all_artifacts=args.remove_all_artifacts,
        config_file=args.config,
    )



"""
Bolt Specification

This module is called bolt to mirror organization of storm package.
"""
from .component import JavaComponentSpec, ShellComponentSpec


class ShellBoltSpec(ShellComponentSpec):
    def __init__(
        self,
        component_cls,
        name=None,
        command=None,
        script=None,
        inputs=None,
        par=1,
        config=None,
        outputs=None,
    ):
        super().__init__(
            component_cls,
            name=name,
            inputs=inputs,
            par=par,
            config=config,
            outputs=outputs,
            command=command,
            script=script,
        )


class JavaBoltSpec(JavaComponentSpec):
    def __init__(
        self,
        component_cls,
        name=None,
        serialized_java=None,
        full_class_name=None,
        args_list=None,
        inputs=None,
        par=1,
        config=None,
        outputs=None,
    ):
        super().__init__(
            component_cls,
            name=name,
            serialized_java=serialized_java,
            inputs=inputs,
            par=par,
            config=config,
            outputs=outputs,
            full_class_name=full_class_name,
            args_list=args_list,
        )



"""
Tests for Topology DSL
"""

import logging
import unittest
from io import BytesIO

from streamparse.dsl import Grouping, Stream, Topology
from streamparse.storm import (
    Bolt,
    Component,
    JavaBolt,
    JavaSpout,
    ShellBolt,
    ShellSpout,
    Spout,
)
from streamparse.thrift import JavaObjectArg


log = logging.getLogger(__name__)


class WordSpout(Spout):
    outputs = ["word"]


class WordCountBolt(Bolt):
    outputs = ["word", "count"]


class MultiStreamWordCountBolt(Bolt):
    outputs = [
        Stream(fields=["word", "count"]),
        Stream(fields=["all_word_count"], name="sum"),
    ]


class DatabaseDumperBolt(Bolt):
    outputs = []


class TopologyTests(unittest.TestCase):
    maxDiff = 1000

    def test_basic_spec(self):
        class WordCount(Topology):
            word_spout = WordSpout.spec(par=2)
            word_bolt = WordCountBolt.spec(
                inputs={word_spout: Grouping.fields("word")}, par=8
            )

        self.assertEqual(len(WordCount.specs), 2)
        self.assertEqual(
            list(WordCount.word_bolt.inputs.keys()), [WordCount.word_spout["default"]]
        )
        self.assertEqual(
            WordCount.thrift_spouts["word_spout"].common.parallelism_hint, 2
        )
        self.assertEqual(WordCount.thrift_bolts["word_bolt"].common.parallelism_hint, 8)
        self.assertEqual(
            WordCount.word_bolt.inputs[WordCount.word_spout["default"]],
            Grouping.fields("word"),
        )

    def test_spec_with_inputs_as_list(self):
        class WordCount(Topology):
            word_spout = WordSpout.spec(par=2)
            word_bolt = WordCountBolt.spec(inputs=[word_spout], par=8)

        self.assertEqual(len(WordCount.specs), 2)
        self.assertEqual(len(WordCount.thrift_spouts), 1)
        self.assertEqual(len(WordCount.thrift_bolts), 1)
        self.assertEqual(
            list(WordCount.word_bolt.inputs.keys()), [WordCount.word_spout["default"]]
        )
        self.assertEqual(
            WordCount.word_bolt.inputs[WordCount.word_spout["default"]],
            Grouping.SHUFFLE,
        )

    def test_multi_stream_bolt(self):
        class WordCount(Topology):
            word_spout = WordSpout.spec(par=2)
            word_bolt = MultiStreamWordCountBolt.spec(inputs=[word_spout], par=8)
            db_dumper_bolt = DatabaseDumperBolt.spec(
                par=4, inputs=[word_bolt["sum"], word_bolt["default"]]
            )

        self.assertEqual(len(WordCount.specs), 3)
        self.assertEqual(len(WordCount.thrift_spouts), 1)
        self.assertEqual(len(WordCount.thrift_bolts), 2)
        self.assertEqual(
            list(WordCount.word_bolt.inputs.keys()), [WordCount.word_spout["default"]]
        )
        self.assertEqual(
            WordCount.word_bolt.inputs[WordCount.word_spout["default"]],
            Grouping.SHUFFLE,
        )
        db_dumper_bolt_input_set = set(WordCount.db_dumper_bolt.inputs.keys())
        self.assertEqual(
            len(WordCount.db_dumper_bolt.inputs.keys()), len(db_dumper_bolt_input_set)
        )
        self.assertEqual(
            db_dumper_bolt_input_set,
            {WordCount.word_bolt["sum"], WordCount.word_bolt["default"]},
        )

    def test_long_chain_spec(self):
        class WordCount(Topology):
            word_spout = WordSpout.spec()
            word_bolt1 = WordCountBolt.spec(inputs=[word_spout])
            word_bolt2 = WordCountBolt.spec(inputs=[word_bolt1])
            word_bolt3 = WordCountBolt.spec(inputs=[word_bolt2])
            word_bolt4 = WordCountBolt.spec(inputs=[word_bolt3])
            word_bolt5 = WordCountBolt.spec(inputs=[word_bolt4])
            word_bolt6 = WordCountBolt.spec(inputs=[word_bolt5])
            word_bolt7 = WordCountBolt.spec(inputs=[word_bolt6])
            word_bolt8 = WordCountBolt.spec(inputs=[word_bolt7])
            word_bolt9 = WordCountBolt.spec(inputs=[word_bolt8])
            word_bolt10 = WordCountBolt.spec(inputs=[word_bolt9])
            word_bolt11 = WordCountBolt.spec(inputs=[word_bolt10])
            word_bolt12 = WordCountBolt.spec(inputs=[word_bolt11])

        self.assertEqual(len(WordCount.specs), 13)
        self.assertEqual(len(WordCount.thrift_spouts), 1)
        self.assertEqual(len(WordCount.thrift_bolts), 12)
        self.assertEqual(
            list(WordCount.word_bolt1.inputs.keys())[0], WordCount.word_spout["default"]
        )
        self.assertEqual(
            WordCount.word_bolt1.inputs[WordCount.word_spout["default"]],
            Grouping.SHUFFLE,
        )
        for i in range(2, 13):
            bolt = getattr(WordCount, f"word_bolt{i}")
            prev_bolt = getattr(WordCount, f"word_bolt{i - 1}")
            self.assertEqual(list(bolt.inputs.keys())[0], prev_bolt["default"])
            self.assertEqual(bolt.inputs[prev_bolt["default"]], Grouping.SHUFFLE)

    def test_many_spouts_spec(self):
        class WordCount(Topology):
            word_spout1 = WordSpout.spec()
            word_spout2 = WordSpout.spec()
            word_spout3 = WordSpout.spec()
            word_spout4 = WordSpout.spec()
            word_spout5 = WordSpout.spec()
            word_bolt = WordCountBolt.spec(
                inputs=[word_spout1, word_spout2, word_spout3, word_spout4, word_spout5]
            )

    def test_invalid_bolt_group_field(self):
        # Fields grouping must specify field output by spout
        with self.assertRaises(ValueError):

            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(
                    inputs={word_spout: Grouping.fields("foo")}
                )

    def test_empty_bolt_group_field(self):
        # Fields groupings require field names to be specified
        with self.assertRaises(ValueError):

            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs={word_spout: Grouping.fields()})

    def test_duplicate_name(self):
        # Each component name must be unique
        with self.assertRaises(ValueError):

            class WordCount(Topology):
                word = WordSpout.spec()
                word_ = WordCountBolt.spec(name="word", inputs=[word])

    def test_no_input_bolt(self):
        # All bolts must have at least one input
        with self.assertRaises(ValueError):

            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs=[])

    def test_no_outputs_spout_empty(self):
        # All spouts must have output fields
        class PointlessSpout(Spout):
            outputs = []

        with self.assertRaises(ValueError):

            class WordCount(Topology):
                word_spout = PointlessSpout.spec()

    def test_no_outputs_spout(self):
        # All spouts must have output fields
        class PointlessSpout(Spout):
            outputs = []

        with self.assertRaises(ValueError):

            class WordCount(Topology):
                word_spout = PointlessSpout.spec()

    def test_base_component_rejection(self):
        # Topology components must inherit from Bolt/Spout, not Component directly
        class MyComponent(Component):
            outputs = []

        with self.assertRaises(TypeError):

            class WordCount(Topology):
                word_spout = MyComponent.spec()

    def test_component_instead_of_spec(self):
        # Make sure we catch things early when people forget to call .spec
        with self.assertRaises(TypeError):

            class WordCount(Topology):
                word_spout = WordSpout(input_stream=BytesIO(), output_stream=BytesIO())

    def test_no_spout(self):
        # Every topology must have a spout
        with self.assertRaises(ValueError):

            class WordCount(Topology):
                pass

    def test_zero_par(self):
        # Component parallelism (number of processes) can't be 0
        with self.assertRaises(ValueError):

            class WordCount(Topology):
                word_spout = WordSpout.spec(par=0)

    def test_negative_par(self):
        # Component parallelism (number of processes) can't be negative
        with self.assertRaises(ValueError):

            class WordCount(Topology):
                word_spout = WordSpout.spec(par=-1)

    def test_float_par(self):
        # Component parallelism (number of processes) must be an int
        with self.assertRaises(TypeError):

            class WordCount(Topology):
                word_spout = WordSpout.spec(par=5.4)

    def test_dict_par(self):
        # Component parallelism (number of processes) can temporarily be a dict
        class WordCount(Topology):
            word_spout = WordSpout.spec(par={"prod": 5, "beta": 1})

        self.assertEqual(
            WordCount.thrift_spouts["word_spout"].common.parallelism_hint,
            {"prod": 5, "beta": 1},
        )

    def test_dict_par_bad_key_type(self):
        # Component parallelism dict must map for str to int
        with self.assertRaises(TypeError):

            class WordCount(Topology):
                word_spout = WordSpout.spec(par={1000: 5, "beta": 1})

    def test_dict_par_bad_value_type(self):
        # Component parallelism dict must map for str to int
        with self.assertRaises(TypeError):

            class WordCount(Topology):
                word_spout = WordSpout.spec(par={"prod": 5.5, "beta": 1})

    def test_invalid_bolt_input_dict_key(self):
        # Keys in input dict must be either GlobalStreamId or ComponentSpec
        with self.assertRaises(TypeError):

            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(
                    inputs={"word_spout": Grouping.fields("word")}
                )

    def test_invalid_bolt_input_dict_val(self):
        # Values in input dict must be Grouping objects
        with self.assertRaises(TypeError):

            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs={word_spout: "word"})

    def test_invalid_bolt_input_str(self):
        # inputs must be either list, dict, or None; not str
        with self.assertRaises(TypeError):

            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs="word_spout")

    def test_invalid_config_str(self):
        # configs must be either dict, or None; not str
        with self.assertRaises(TypeError):

            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(
                    inputs=[word_spout], config='{"foo": "bar"}'
                )

    def test_invalid_topology_config_str(self):
        # configs must be either dict, or None; not str
        with self.assertRaises(TypeError):

            class WordCount(Topology):
                config = '{"foo": "bar"}'
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs=[word_spout])

    def test_invalid_outputs_entry_int(self):
        # Outputs must either be list of strings or list of streams, not ints
        class BadSpout(Spout):
            outputs = ["foo", 2]

        with self.assertRaises(TypeError):

            class WordCount(Topology):
                bad_spout = BadSpout.spec()
                word_bolt = WordCountBolt.spec(inputs=[bad_spout])

    def test_invalid_outputs_str(self):
        # Outputs must either be list of strings or list of streams
        class BadSpout(Spout):
            outputs = "foo"

        with self.assertRaises(TypeError):

            class WordCount(Topology):
                bad_spout = BadSpout.spec()
                word_bolt = WordCountBolt.spec(inputs=[bad_spout])

    def test_unknown_stream(self):
        # Should raise keyerror if stream name is not valid
        with self.assertRaises(KeyError):

            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs=[word_spout["word"]])

    def test_perl_bolt(self):
        # Make sure ShellBolt.spec works
        class PerlWordCount(Topology):
            word_spout = WordSpout.spec(par=2)
            word_bolt = ShellBolt.spec(
                command="perl",
                script="count_words.pl",
                inputs={word_spout: Grouping.fields("word")},
                par=8,
                outputs=["word", "count"],
            )

        self.assertEqual(len(PerlWordCount.specs), 2)
        self.assertEqual(
            list(PerlWordCount.word_bolt.inputs.keys()),
            [PerlWordCount.word_spout["default"]],
        )
        self.assertEqual(
            PerlWordCount.thrift_spouts["word_spout"].common.parallelism_hint, 2
        )
        self.assertEqual(
            PerlWordCount.thrift_bolts["word_bolt"].common.parallelism_hint, 8
        )
        self.assertEqual(
            PerlWordCount.word_bolt.inputs[PerlWordCount.word_spout["default"]],
            Grouping.fields("word"),
        )

    def test_shell_bolt_no_command(self):
        # Should raise ValueError if command is None
        with self.assertRaises(ValueError):

            class PerlWordCount(Topology):
                word_spout = WordSpout.spec(par=2)
                word_bolt = ShellBolt.spec(
                    script="count_words.pl",
                    inputs={word_spout: Grouping.fields("word")},
                    par=8,
                    outputs=["word", "count"],
                )

    def test_shell_bolt_no_script(self):
        # Should raise TypeError if script is None
        with self.assertRaises(TypeError):

            class PerlWordCount(Topology):
                word_spout = WordSpout.spec(par=2)
                word_bolt = ShellBolt.spec(
                    command="perl",
                    inputs={word_spout: Grouping.fields("word")},
                    par=8,
                    outputs=["word", "count"],
                )

    def test_java_bolt_valid_arg_list(self):
        # JavaBolt should work fine when given basic data types
        class JavaWordCount(Topology):
            word_spout = WordSpout.spec(par=2)
            word_bolt = JavaBolt.spec(
                full_class_name="com.bar.foo.counter.WordCountBolt",
                args_list=["foo", 1, b"\x09\x10", True, 3.14159],
                inputs={word_spout: Grouping.fields("word")},
                par=8,
                outputs=["word", "count"],
            )

        java_object = JavaWordCount.thrift_bolts["word_bolt"].bolt_object.java_object
        self.assertEqual(
            java_object.full_class_name, "com.bar.foo.counter.WordCountBolt"
        )
        self.assertEqual(
            java_object.args_list,
            [
                JavaObjectArg(string_arg="foo"),
                JavaObjectArg(long_arg=1),
                JavaObjectArg(binary_arg=b"\x09\x10"),
                JavaObjectArg(bool_arg=True),
                JavaObjectArg(double_arg=3.14159),
            ],
        )
        self.assertIsNone(
            JavaWordCount.thrift_bolts["word_bolt"].bolt_object.serialized_java
        )
        self.assertIsNone(JavaWordCount.thrift_bolts["word_bolt"].bolt_object.shell)

    def test_java_bolt_invalid_arg_list(self):
        # JavaBolt should raise TypeError when given something other than basic data types
        with self.assertRaises(TypeError):

            class JavaWordCount(Topology):
                word_spout = WordSpout.spec(par=2)
                word_bolt = JavaBolt.spec(
                    full_class_name="com.bar.foo.counter.WordCountBolt",
                    args_list=[{"foo": "bar"}, 1],
                    inputs={word_spout: Grouping.fields("word")},
                    par=8,
                    outputs=["word", "count"],
                )

    def test_java_bolt_valid_serialized_java(self):
        # JavaBolt should work fine when given a byte string for serialized_java
        serialized_java = b"\x01\x02\x03\x04\x05"

        class JavaWordCount(Topology):
            word_spout = WordSpout.spec(par=2)
            word_bolt = JavaBolt.spec(
                serialized_java=serialized_java,
                inputs={word_spout: Grouping.fields("word")},
                par=8,
                outputs=["word", "count"],
            )

        self.assertEqual(
            JavaWordCount.thrift_bolts["word_bolt"].bolt_object.serialized_java,
            serialized_java,
        )
        self.assertIsNone(
            JavaWordCount.thrift_bolts["word_bolt"].bolt_object.java_object
        )
        self.assertIsNone(JavaWordCount.thrift_bolts["word_bolt"].bolt_object.shell)

    def test_perl_spout(self):
        # Make sure ShellBolt.spec works
        class PerlWordCount(Topology):
            word_spout = ShellSpout.spec(
                command="perl", script="count_words.pl", par=2, outputs=["word"]
            )
            word_bolt = WordCountBolt.spec(
                inputs={word_spout: Grouping.fields("word")}, par=8
            )

        self.assertEqual(len(PerlWordCount.specs), 2)
        self.assertEqual(
            list(PerlWordCount.word_bolt.inputs.keys()),
            [PerlWordCount.word_spout["default"]],
        )
        self.assertEqual(
            PerlWordCount.thrift_spouts["word_spout"].common.parallelism_hint, 2
        )
        self.assertEqual(
            PerlWordCount.thrift_bolts["word_bolt"].common.parallelism_hint, 8
        )
        self.assertEqual(
            PerlWordCount.word_bolt.inputs[PerlWordCount.word_spout["default"]],
            Grouping.fields("word"),
        )

    def test_shell_spout_no_command(self):
        # Should raise ValueError if command is None
        with self.assertRaises(ValueError):

            class PerlWordCount(Topology):
                word_spout = ShellSpout.spec(
                    script="count_words.pl", par=8, outputs=["word"]
                )
                word_bolt = WordCountBolt.spec(
                    inputs={word_spout: Grouping.fields("word")}, par=8
                )

    def test_shell_spout_no_script(self):
        # Should raise TypeError if script is None
        with self.assertRaises(TypeError):

            class PerlWordCount(Topology):
                word_spout = ShellSpout.spec(command="perl", par=8, outputs=["word"])
                word_bolt = WordCountBolt.spec(
                    inputs={word_spout: Grouping.fields("word")}, par=8
                )

    def test_java_spout_valid_arg_list(self):
        # JavaSpout should work fine when given basic data types
        class JavaWordCount(Topology):
            word_spout = JavaSpout.spec(
                full_class_name="com.bar.foo.counter.WordSpout",
                args_list=["foo", 1, b"\x09\x10", True, 3.14159],
                par=8,
                outputs=["word"],
            )
            word_bolt = WordCountBolt.spec(
                inputs={word_spout: Grouping.fields("word")}, par=8
            )

        java_object = JavaWordCount.thrift_spouts["word_spout"].spout_object.java_object
        self.assertEqual(java_object.full_class_name, "com.bar.foo.counter.WordSpout")
        self.assertEqual(
            java_object.args_list,
            [
                JavaObjectArg(string_arg="foo"),
                JavaObjectArg(long_arg=1),
                JavaObjectArg(binary_arg=b"\x09\x10"),
                JavaObjectArg(bool_arg=True),
                JavaObjectArg(double_arg=3.14159),
            ],
        )
        self.assertIsNone(
            JavaWordCount.thrift_spouts["word_spout"].spout_object.serialized_java
        )
        self.assertIsNone(JavaWordCount.thrift_spouts["word_spout"].spout_object.shell)

    def test_java_spout_invalid_arg_list(self):
        # JavaSpout should raise TypeError when given something other than basic data types
        with self.assertRaises(TypeError):

            class JavaWordCount(Topology):
                word_spout = JavaSpout.spec(
                    full_class_name="com.bar.foo.counter.WordSpout",
                    args_list=[{"foo": "bar"}, 1],
                    par=8,
                    outputs=["word"],
                )
                word_bolt = WordCountBolt.spec(
                    inputs={word_spout: Grouping.fields("word")}, par=8
                )

    def test_java_spout_valid_serialized_java(self):
        # JavaSpout should work fine when given a byte string for serialized_java
        serialized_java = b"\x01\x02\x03\x04\x05"

        class JavaWordCount(Topology):
            word_spout = JavaSpout.spec(
                serialized_java=serialized_java, par=8, outputs=["word"]
            )
            word_bolt = WordCountBolt.spec(
                inputs={word_spout: Grouping.fields("word")}, par=8
            )

        self.assertEqual(
            JavaWordCount.thrift_spouts["word_spout"].spout_object.serialized_java,
            serialized_java,
        )
        self.assertIsNone(
            JavaWordCount.thrift_spouts["word_spout"].spout_object.java_object
        )
        self.assertIsNone(JavaWordCount.thrift_spouts["word_spout"].spout_object.shell)

    def test_basic_to_flux_dict(self):
        class WordCount(Topology):
            word_spout = WordSpout.spec(par=2)
            word_bolt = WordCountBolt.spec(
                inputs={word_spout: Grouping.fields("word")}, par=8
            )

        self.assertEqual(
            WordCount.to_flux_dict("word_count"),
            {
                "name": "word_count",
                "spouts": [
                    {
                        "id": "word_spout",
                        "className": "org.apache.storm.flux.wrappers.spouts.FluxShellSpout",
                        "configMethods": [],
                        "constructorArgs": [
                            ["streamparse_run", "test.streamparse.test_dsl.WordSpout"],
                            ["word"],
                        ],
                        "parallelism": 2,
                    }
                ],
                "streams": [
                    {
                        "to": "word_bolt",
                        "from": "word_spout",
                        "grouping": {
                            "streamId": "default",
                            "args": ["word"],
                            "type": "FIELDS",
                        },
                    }
                ],
                "bolts": [
                    {
                        "id": "word_bolt",
                        "className": "org.apache.storm.flux.wrappers.bolts.FluxShellBolt",
                        "configMethods": [],
                        "constructorArgs": [
                            [
                                "streamparse_run",
                                "test.streamparse.test_dsl.WordCountBolt",
                            ],
                            ["word", "count"],
                        ],
                        "parallelism": 8,
                    }
                ],
            },
        )



"""
Utility functions for the DSL
"""

from ..thrift import JavaObjectArg


def to_java_arg(arg):
    """Converts Python objects to equivalent Thrift JavaObjectArgs"""
    if isinstance(arg, bool):
        java_arg = JavaObjectArg(bool_arg=arg)
    elif isinstance(arg, int):
        # Just use long all the time since Python 3 doesn't
        # distinguish between long and int
        java_arg = JavaObjectArg(long_arg=arg)
    elif isinstance(arg, bytes):
        java_arg = JavaObjectArg(binary_arg=arg)
    elif isinstance(arg, str):
        java_arg = JavaObjectArg(string_arg=arg)
    elif isinstance(arg, float):
        java_arg = JavaObjectArg(double_arg=arg)
    else:
        raise TypeError(
            "Only basic data types can be specified"
            " as arguments to JavaObject "
            "constructors.  Given: {!r}".format(arg)
        )
    return java_arg


def to_python_arg(java_arg):
    """Convert a Thrift JavaObjectArg into a basic Python data type"""
    arg = None
    for val in java_arg.__dict__.values():
        if val is not None:
            arg = val
            break
    return arg


def to_python_arg_list(java_arg_list):
    """Convert a list of Thrift JavaObjectArg objects into a list of basic types"""
    return [to_python_arg(java_arg) for java_arg in java_arg_list]



"""
Create a JAR that can be used to deploy a topology to Storm.
"""

import os
import shutil
import sys

from fabric.api import hide, local, settings

from ..util import prepare_topology
from .common import add_simple_jar


def jar_for_deploy(simple_jar=False):
    """ Build a jar to use for deploying the topology. """
    # Create _resources folder which will contain Python code in JAR
    prepare_topology()
    # Use Leiningen to clean up and build JAR
    jar_type = "JAR" if simple_jar else "Uber-JAR"
    print("Cleaning from prior builds...")
    sys.stdout.flush()
    with hide("running", "stdout"):
        res = local("lein clean")
    if not res.succeeded:
        raise RuntimeError(
            f"Unable to run 'lein clean'!\nSTDOUT:\n{res.stdout}\nSTDERR:\n{res.stderr}"
        )
    print(f"Creating topology {jar_type}...")
    sys.stdout.flush()
    cmd = "lein jar" if simple_jar else "lein uberjar"
    with hide("running"), settings(warn_only=True):
        res = local(cmd, capture=True)
        if not res.succeeded:
            raise RuntimeError(
                f"Unable to run '{cmd}'!\nSTDOUT:\n{res.stdout}\nSTDERR:\n{res.stderr}"
            )
    # XXX: This will fail if more than one JAR is built
    lines = res.stdout.splitlines()
    for line in lines:
        line = line.strip()
        if not line.startswith("Created"):
            continue
        line = line.replace("Created ", "")
        # != is XOR
        if simple_jar != line.endswith("standalone.jar"):
            jar = line
            break
    else:
        raise RuntimeError(
            "Failed to find JAR in '{}' output\nSTDOUT:\n{}"
            "STDERR:\n{}".format(cmd, res.stdout, res.stderr)
        )
    print(f"{jar_type} created: {jar}")
    sys.stdout.flush()
    print("Removing _resources temporary directory...", end="")
    sys.stdout.flush()
    resources_dir = os.path.join("_resources", "resources")
    if os.path.isdir(resources_dir):
        shutil.rmtree(resources_dir)
    print("done")
    return jar


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser("jar", description=__doc__, help=main.__doc__)
    subparser.set_defaults(func=main)
    add_simple_jar(subparser)


def main(args):
    """ Create a deployable JAR for a topology. """
    jar_for_deploy(simple_jar=args.simple_jar)



"""
Test target for streamparse_run
"""

from test.streamparse.test_streamparse_run import StreamparseRunTests


class StreamparseRunTarget:
    def __init__(self, serializer):
        self.serializer = serializer

    def run(self):
        StreamparseRunTests.run_target_invoked = True
        StreamparseRunTests.run_target_invoked_serializer = self.serializer



import argparse
import unittest

from streamparse.cli import sparse


def test_load_subparsers():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    sparse.load_subparsers(subparsers)
    # grab subcommands from subparsers
    subcommands = parser._optionals._actions[1].choices.keys()
    # we know quickstart will be a subcommand test others as needed
    assert "quickstart" in subcommands



"""
Component-level Specification

This module is called component to mirror organization of pystorm package.
"""
from copy import deepcopy

import simplejson as json

from ..thrift import (
    ComponentCommon,
    ComponentObject,
    GlobalStreamId,
    JavaObject,
    ShellComponent,
    storm_thrift,
    StreamInfo,
)
from .stream import Grouping, Stream
from .util import to_java_arg


class ComponentSpec:
    """Describes the inputs, outputs, etc. for a Storm topology component.

    Generates some of the Thrift data structures that are needed to build a
    Storm topology.
    """

    def __init__(
        self, component_cls, name=None, inputs=None, par=1, config=None, outputs=None
    ):
        self.component_cls = component_cls
        self.name = name
        self.par = self._sanitize_par(component_cls, par)
        self.config = self._sanitize_config(component_cls, config)
        self.outputs = self._sanitize_outputs(component_cls, outputs)
        self.inputs = self._sanitize_inputs(inputs)
        self.common = ComponentCommon(
            inputs=self.inputs,
            streams=self.outputs,
            parallelism_hint=self.par,
            json_conf=self.config,
        )

    @staticmethod
    def _sanitize_par(component_cls, par):
        """ Raises exceptions if `par` value is not a positive integer. """
        if par is None:
            par = component_cls.par
        if isinstance(par, dict):
            for stage, par_hint in par.items():
                if not (isinstance(stage, str) and isinstance(par_hint, int)):
                    raise TypeError(
                        "If par is a dict, it must map from "
                        "environment names to integers specifying "
                        "the parallelism hint for the component.\n"
                        "Given stage: {!r}\n"
                        "Given parallelism hint: {!r}".format(stage, par_hint)
                    )
                elif par_hint < 1:
                    raise ValueError(
                        "Parallelism hint for stage {} must be an "
                        "integer greater than 0. Given: {}".format(stage, par_hint)
                    )
        elif not isinstance(par, int):
            raise TypeError(
                f"Parallelism hint must be an integer greater than 0. Given: {par!r}"
            )
        elif par < 1:
            raise ValueError(
                f"Parallelism hint must be an integer greater than 0. Given: {par}"
            )
        return par

    @staticmethod
    def _sanitize_inputs(inputs):
        if isinstance(inputs, dict):
            for key, val in list(inputs.items()):
                if not isinstance(key, GlobalStreamId):
                    if isinstance(key, ComponentSpec):
                        inputs[key["default"]] = val
                        del inputs[key]
                    else:
                        raise TypeError(
                            "If inputs is a dict, it is expected to"
                            " map from GlobalStreamId or "
                            "ComponentSpec objects to Grouping "
                            "objects.  Given key: {!r}; Given "
                            "value: {!r}".format(key, val)
                        )
                if not isinstance(val, storm_thrift.Grouping):
                    raise TypeError(
                        "If inputs is a dict, it is expected to map"
                        " from GlobalStreamId or ComponentSpec "
                        "objects to Grouping objects.  Given key: "
                        "{!r}; Given value: {!r}".format(key, val)
                    )
            input_dict = inputs
        else:
            if isinstance(inputs, ComponentSpec):
                inputs = [inputs]
            if isinstance(inputs, (list, tuple)):
                input_dict = {}
                for input_spec in inputs:
                    grouping = Grouping.SHUFFLE
                    if isinstance(input_spec, ComponentSpec):
                        component_id = input_spec.name or input_spec
                        stream_id = GlobalStreamId(
                            componentId=component_id, streamId="default"
                        )
                        # Can only automatically determine if grouping should be
                        # direct when given a ComponentSpec.  If
                        # GlobalStreamId, we're out of luck.
                        # TODO: Document this.
                        default_stream = input_spec.common.streams.get("default")
                        if default_stream is not None and default_stream.direct:
                            grouping = Grouping.DIRECT
                    elif isinstance(input_spec, GlobalStreamId):
                        stream_id = input_spec
                    else:
                        raise TypeError(
                            "Inputs must be ComponentSpec or "
                            "GlobalStreamId objects.  Given: {!r}".format(input_spec)
                        )
                    input_dict[stream_id] = grouping
            elif inputs is None:
                input_dict = {}
            else:
                raise TypeError(
                    f"Inputs must either be a list, dict, or None.  Given: {inputs!r}"
                )
        return input_dict

    @staticmethod
    def _sanitize_config(component_cls, config):
        if config is None:
            config = component_cls.config
        if isinstance(config, dict):
            config = json.dumps(config)
        elif config is None:
            config = "{}"
        else:
            raise TypeError(
                f"Config must either be a dict or None.  Given: {config!r}"
            )
        return config

    @staticmethod
    def _sanitize_outputs(component_cls, outputs):
        if outputs is None:
            outputs = component_cls.outputs
        if outputs is None:
            outputs = []
        if isinstance(outputs, (list, tuple)):
            streams = {}
            for output in outputs:
                if isinstance(output, Stream):
                    streams[output.name] = StreamInfo(
                        output_fields=output.fields, direct=output.direct
                    )
                # Strings are output fields for default stream
                elif isinstance(output, str):
                    default = streams.setdefault(
                        "default", StreamInfo(output_fields=[], direct=False)
                    )
                    default.output_fields.append(output)
                else:
                    raise TypeError(
                        "Outputs must either be a list of strings "
                        "or a list of Streams.  Invalid entry: {!r}".format(output)
                    )
        else:
            raise TypeError(
                "Outputs must either be a list of strings or a list"
                " of Streams.  Given: {!r}".format(outputs)
            )
        return streams

    def __getitem__(self, stream):
        if stream not in self.common.streams:
            raise KeyError(
                "Invalid stream for {}: {!r}. Valid streams are: "
                "{}".format(self.name, stream, list(self.common.streams.keys()))
            )
        # If name is None, because it hasn't been set yet, use object instead
        component_id = self.name or self
        return GlobalStreamId(componentId=component_id, streamId=stream)

    def __repr__(self):
        """:returns: A string representation of the Specification. """
        attr_dict = deepcopy(self.__dict__)
        component_cls = attr_dict.pop("component_cls")
        repr_str = f"{self.__class__.__name__}({component_cls.__name__}"
        for key, val in attr_dict.items():
            repr_str += f", {key}={val!r}"
        repr_str += ")"
        return repr_str


class JavaComponentSpec(ComponentSpec):
    """ComponentSpec for JVM-based topology components."""

    def __init__(
        self,
        component_cls,
        name=None,
        serialized_java=None,
        full_class_name=None,
        args_list=None,
        inputs=None,
        par=None,
        config=None,
        outputs=None,
    ):
        super().__init__(
            component_cls,
            name=name,
            inputs=inputs,
            par=par,
            config=config,
            outputs=outputs,
        )
        if serialized_java is not None:
            if isinstance(serialized_java, bytes):
                comp_object = ComponentObject(serialized_java=serialized_java)
                self.component_object = comp_object
            else:
                raise TypeError("serialized_java must be either bytes or None")
        else:
            if not full_class_name:
                raise ValueError("full_class_name is required")
            if args_list is None:
                raise TypeError("args_list must not be None")
            else:
                # Convert arguments to JavaObjectArgs
                for i, arg in enumerate(args_list):
                    args_list[i] = to_java_arg(arg)
            java_object = JavaObject(
                full_class_name=full_class_name, args_list=args_list
            )
            self.component_object = ComponentObject(java_object=java_object)


class ShellComponentSpec(ComponentSpec):
    """ComponentSpec for shell-based topology components (like Python ones)."""

    def __init__(
        self,
        component_cls,
        name=None,
        command=None,
        script=None,
        inputs=None,
        par=1,
        config=None,
        outputs=None,
    ):
        super().__init__(
            component_cls,
            name=name,
            inputs=inputs,
            par=par,
            config=config,
            outputs=outputs,
        )
        if not command:
            raise ValueError("command is required")
        if script is None:
            raise TypeError(
                "script must not be None.  If your command does not"
                " take arguments, specify the empty string for "
                "script."
            )
        shell_component = ShellComponent(execution_command=command, script=script)
        self.component_object = ComponentObject(shell=shell_component)



"""
Display slots used by every topology on the cluster
"""

from collections import Counter, defaultdict

from pkg_resources import parse_version

from .common import add_config, add_environment
from ..util import get_ui_json, get_ui_jsons, print_stats_table, storm_lib_version


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser(
        "slot_usage", description=__doc__, help=main.__doc__
    )
    subparser.set_defaults(func=main)
    add_config(subparser)
    add_environment(subparser)


def display_slot_usage(env_name, config_file=None):
    print(
        "Querying Storm UI REST service for slot usage stats (this can take a while)..."
    )
    topology_summary_path = "/api/v1/topology/summary"
    topology_detail_path = "/api/v1/topology/{topology}"
    component_path = "/api/v1/topology/{topology}/component/{component}"
    topo_summary_json = get_ui_json(
        env_name, topology_summary_path, config_file=config_file
    )
    topology_ids = [x["id"] for x in topo_summary_json["topologies"]]
    # Keep track of the number of workers used by each topology on each machine
    topology_worker_ports = defaultdict(lambda: defaultdict(set))
    topology_executor_counts = defaultdict(Counter)
    topology_names = set()
    topology_components = dict()
    topology_detail_jsons = get_ui_jsons(
        env_name,
        (topology_detail_path.format(topology=topology) for topology in topology_ids),
        config_file=config_file,
    )

    for topology in topology_ids:
        topology_detail_json = topology_detail_jsons[
            topology_detail_path.format(topology=topology)
        ]
        spouts = [x["spoutId"] for x in topology_detail_json["spouts"]]
        bolts = [x["boltId"] for x in topology_detail_json["bolts"]]
        topology_components[topology] = spouts + bolts

    comp_details = get_ui_jsons(
        env_name,
        (
            component_path.format(topology=topology, component=comp)
            for topology, comp_list in topology_components.items()
            for comp in comp_list
        ),
        config_file=config_file,
    )

    for request_url, comp_detail in comp_details.items():
        topology = request_url.split("/")[4]
        topology_detail_json = topology_detail_jsons[
            topology_detail_path.format(topology=topology)
        ]
        for worker in comp_detail["executorStats"]:
            topology_worker_ports[worker["host"]][topology_detail_json["name"]].add(
                worker["port"]
            )
            topology_executor_counts[worker["host"]][topology_detail_json["name"]] += 1
            topology_names.add(topology_detail_json["name"])

    topology_names = sorted(topology_names)
    columns = ["Host"] + topology_names
    rows = [
        (
            [host]
            + [
                "{} ({})".format(
                    len(host_dict.get(topology, set())),
                    topology_executor_counts[host][topology],
                )
                for topology in topology_names
            ]
        )
        for host, host_dict in sorted(topology_worker_ports.items())
    ]
    print_stats_table("Slot (and Executor) Counts by Topology", rows, columns)


def main(args):
    """ Display slots used by every topology on the cluster. """
    storm_version = storm_lib_version()
    if storm_version >= parse_version("0.9.2-incubating"):
        display_slot_usage(args.environment)
    else:
        print(f"ERROR: Storm {storm_version} does not support this command.")



"""
Utilities for bootstrapping streamparse projects.
"""
from contextlib import contextmanager
import os
import sys
import shutil

import pkg_resources
from fabric.colors import green, red, blue
from jinja2 import Environment, FileSystemLoader


_path_prefixes = []
_path_prefix = ""
_root = os.path.abspath(os.path.dirname(__file__))


def _here(*paths):
    path = os.path.join(*paths)
    filename = pkg_resources.resource_filename(__name__, path)
    return filename


_env = Environment(loader=FileSystemLoader(_here("project")))


@contextmanager
def _cd(path):
    global _path_prefixes
    global _path_prefix
    _path_prefixes.append(path)
    _path_prefix = "/".join(_path_prefixes)
    yield
    _path_prefixes.pop()
    _path_prefix = "/".join(_path_prefixes)


def _mkdir(path):
    path = f"{_path_prefix}/{path}" if _path_prefix != "" else path
    print(f"    {green('create'):<18} {path}")
    os.makedirs(path)


def _cp(src, dest):
    dest = f"{_path_prefix}/{dest}" if _path_prefix != "" else dest
    print(f"    {green('create'):<18} {dest}")
    shutil.copy(src, dest)


def _touch(filename):
    filename = (
        f"{_path_prefix}/{filename}" if _path_prefix != "" else filename
    )
    print(f"    {green('create'):<18} {filename}")
    with open(filename, "w"):
        pass


def _generate(template_filename, dest):
    dest = f"{_path_prefix}/{dest}" if _path_prefix != "" else dest
    print(f"    {green('create'):<18} {dest}")
    template = _env.get_template(template_filename)
    with open(dest, "w") as fp:
        fp.write(template.render())


def quickstart(project_name):
    # TODO: alternate way maybe to do all of this is do something like
    # glob.glob('project/**/*') and then we copy everything that's doesn't have
    # jinja2 in filename, generate the jinja2 stuff
    if os.path.exists(project_name):
        print(f"{red('error')}: folder \"{project_name}\" already exists")
        sys.exit(1)

    print(f"\nCreating your {blue(project_name)} streamparse project...")
    _env.globals["project_name"] = project_name

    _mkdir(project_name)
    with _cd(project_name):
        _cp(_here("project", "gitignore"), ".gitignore")
        _generate("config.jinja2.json", "config.json")
        _cp(_here("project", "fabfile.py"), "fabfile.py")
        _generate("project.jinja2.clj", "project.clj")
        _touch("README.md")
        _mkdir("src")
        with _cd("src"):
            _mkdir("bolts")
            with _cd("bolts"):
                _cp(_here("project", "src", "bolts", "__init__.py"), "__init__.py")
                _cp(_here("project", "src", "bolts", "wordcount.py"), "wordcount.py")
            _mkdir("spouts")
            with _cd("spouts"):
                _cp(_here("project", "src", "spouts", "__init__.py"), "__init__.py")
                _cp(_here("project", "src", "spouts", "words.py"), "words.py")
        _mkdir("topologies")
        with _cd("topologies"):
            _cp(_here("project", "topologies", "wordcount.py"), "wordcount.py")
        _mkdir("virtualenvs")
        with _cd("virtualenvs"):
            _cp(_here("project", "virtualenvs", "wordcount.txt"), "wordcount.txt")

    print("Done.\n")
    print(
        (
            "Try running your topology locally with:\n\n" "\tcd {}\n" "\tsparse run"
        ).format(project_name)
    )



"""
Create new streamparse project template.
"""

from streamparse.bootstrap import quickstart


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser(
        "quickstart", description=__doc__, help=main.__doc__
    )
    subparser.set_defaults(func=main)
    subparser.add_argument("project_name", help="Name of new streamparse project.")


def main(args):
    """ Create new streamparse project template. """
    quickstart(args.project_name)



import argparse
import unittest
from unittest.mock import patch

from streamparse.cli.run import main, subparser_hook


def test_subparser_hook():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    subparser_hook(subparsers)

    subcommands = parser._optionals._actions[1].choices.keys()
    assert "run" in subcommands


@patch("streamparse.cli.run.run_local_topology", autospec=True)
def test_main_args_passed(run_local_mock):
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    subparser_hook(subparsers)

    args = parser.parse_args("run -e my_env -n my_topo --ackers 1".split())

    main(args)
    run_local_mock.assert_called_with(
        name="my_topo",
        options={"topology.acker.executors": 1},
        env_name="my_env",
        time=0,
        config_file=None,
    )



"""
Submit a Storm topology to Nimbus.
"""

import importlib
import os
import sys
import time
from itertools import chain

import simplejson as json
from fabric.api import env
from pkg_resources import parse_version

from ..dsl.component import JavaComponentSpec
from ..thrift import ShellComponent, SubmitOptions, TopologyInitialStatus

from ..util import (
    activate_env,
    get_config,
    get_env_config,
    get_nimbus_client,
    get_topology_definition,
    get_topology_from_file,
    nimbus_storm_version,
    set_topology_serializer,
    ssh_tunnel,
    warn,
)
from .common import (
    add_ackers,
    add_config,
    add_debug,
    add_environment,
    add_name,
    add_options,
    add_override_name,
    add_overwrite_virtualenv,
    add_pool_size,
    add_requirements,
    add_timeout,
    add_user,
    add_wait,
    add_workers,
    resolve_options,
    warn_about_deprecated_user,
)
from .jar import jar_for_deploy
from .kill import _kill_topology
from .list import _list_topologies
from .update_virtualenv import create_or_update_virtualenvs


THRIFT_CHUNK_SIZE = 307200


def get_user_tasks():
    """Get tasks defined in a user's tasks.py and fabfile.py file which is
    assumed to be in the current working directory.

    :returns: a `tuple` (invoke_tasks, fabric_tasks)
    """
    sys.path.insert(0, os.getcwd())
    try:
        user_invoke = importlib.import_module("tasks")
    except ImportError:
        user_invoke = None
    try:
        user_fabric = importlib.import_module("fabfile")
    except ImportError:
        user_fabric = None
    return user_invoke, user_fabric


def is_safe_to_submit(topology_name, nimbus_client):
    """Is topology not in list of current topologies?"""
    topologies = _list_topologies(nimbus_client)
    safe = not any(topology.name == topology_name for topology in topologies)
    return safe


def _kill_existing_topology(topology_name, force, wait, nimbus_client):
    if force and not is_safe_to_submit(topology_name, nimbus_client):
        print(f'Killing current "{topology_name}" topology.')
        sys.stdout.flush()
        _kill_topology(topology_name, nimbus_client, wait=wait)
        while not is_safe_to_submit(topology_name, nimbus_client):
            print(f"Waiting for topology {topology_name} to quit...")
            sys.stdout.flush()
            time.sleep(0.5)
        print("Killed.")
        sys.stdout.flush()


def _submit_topology(
    topology_name,
    topology_class,
    remote_jar_path,
    config,
    env_config,
    nimbus_client,
    options=None,
    active=True,
):
    if options.get("pystorm.log.path"):
        print(f"Routing Python logging to {options['pystorm.log.path']}.")
        sys.stdout.flush()

    set_topology_serializer(env_config, config, topology_class)

    # Check if topology name is okay on Storm versions that support that
    if nimbus_storm_version(nimbus_client) >= parse_version("1.1.0"):
        if not nimbus_client.isTopologyNameAllowed(topology_name):
            raise ValueError(
                f"Nimbus says {topology_name} is an invalid name for a Storm topology."
            )

    print(f"Submitting {topology_name} topology to nimbus...", end="")
    sys.stdout.flush()
    initial_status = (
        TopologyInitialStatus.ACTIVE if active else TopologyInitialStatus.INACTIVE
    )
    submit_options = SubmitOptions(initial_status=initial_status)
    nimbus_client.submitTopologyWithOpts(
        name=topology_name,
        uploadedJarLocation=remote_jar_path,
        jsonConf=json.dumps(options),
        topology=topology_class.thrift_topology,
        options=submit_options,
    )
    print("done")


def _pre_submit_hooks(topology_name, env_name, env_config, options):
    """Pre-submit hooks for invoke and fabric."""
    user_invoke, user_fabric = get_user_tasks()
    pre_submit_invoke = getattr(user_invoke, "pre_submit", None)
    if callable(pre_submit_invoke):
        pre_submit_invoke(topology_name, env_name, env_config, options)
    pre_submit_fabric = getattr(user_fabric, "pre_submit", None)
    if callable(pre_submit_fabric):
        pre_submit_fabric(topology_name, env_name, env_config, options)


def _post_submit_hooks(topology_name, env_name, env_config, options):
    """Post-submit hooks for invoke and fabric."""
    user_invoke, user_fabric = get_user_tasks()
    post_submit_invoke = getattr(user_invoke, "post_submit", None)
    if callable(post_submit_invoke):
        post_submit_invoke(topology_name, env_name, env_config, options)
    post_submit_fabric = getattr(user_fabric, "post_submit", None)
    if callable(post_submit_fabric):
        post_submit_fabric(topology_name, env_name, env_config, options)


def _upload_jar(nimbus_client, local_path):
    upload_location = nimbus_client.beginFileUpload()
    print(
        f"Uploading topology jar {local_path} to assigned location: {upload_location}"
    )
    total_bytes = os.path.getsize(local_path)
    bytes_uploaded = 0
    with open(local_path, "rb") as local_jar:
        while True:
            print(f"Uploaded {bytes_uploaded}/{total_bytes} bytes", end="\r")
            sys.stdout.flush()
            curr_chunk = local_jar.read(THRIFT_CHUNK_SIZE)
            if not curr_chunk:
                break
            nimbus_client.uploadChunk(upload_location, curr_chunk)
            bytes_uploaded += len(curr_chunk)
        nimbus_client.finishFileUpload(upload_location)
        print(f"Uploaded {bytes_uploaded}/{total_bytes} bytes")
        sys.stdout.flush()
    return upload_location


def submit_topology(
    name=None,
    env_name=None,
    options=None,
    force=False,
    wait=None,
    simple_jar=True,
    override_name=None,
    requirements_paths=None,
    local_jar_path=None,
    remote_jar_path=None,
    timeout=None,
    config_file=None,
    overwrite_virtualenv=False,
    user=None,
    active=True,
):
    """Submit a topology to a remote Storm cluster."""
    warn_about_deprecated_user(user, "submit_topology")
    config = get_config(config_file=config_file)
    name, topology_file = get_topology_definition(name, config_file=config_file)
    env_name, env_config = get_env_config(env_name, config_file=config_file)
    topology_class = get_topology_from_file(topology_file)
    if override_name is None:
        override_name = name
    if remote_jar_path and local_jar_path:
        warn("Ignoring local_jar_path because given remote_jar_path")
        local_jar_path = None

    # Setup the fabric env dictionary
    activate_env(env_name)

    # Handle option conflicts
    options = resolve_options(options, env_config, topology_class, override_name)

    # Check if we need to maintain virtualenv during the process
    use_venv = options.get("use_virtualenv", True)

    # Check if user wants to install virtualenv during the process
    install_venv = options.get("install_virtualenv", use_venv)

    # Run pre_submit actions provided by project
    _pre_submit_hooks(override_name, env_name, env_config, options)

    # If using virtualenv, set it up, and make sure paths are correct in specs
    if use_venv:
        virtualenv_name = options.get("virtualenv_name", override_name)
        if install_venv:
            create_or_update_virtualenvs(
                env_name,
                name,
                options,
                virtualenv_name=virtualenv_name,
                requirements_paths=requirements_paths,
                config_file=config_file,
                overwrite_virtualenv=overwrite_virtualenv,
                user=user,
            )
        streamparse_run_path = "/".join(
            [env.virtualenv_root, virtualenv_name, "bin", "streamparse_run"]
        )
        # Update python paths in bolts
        for thrift_bolt in topology_class.thrift_bolts.values():
            inner_shell = thrift_bolt.bolt_object.shell
            if isinstance(inner_shell, ShellComponent):
                if "streamparse_run" in inner_shell.execution_command:
                    inner_shell.execution_command = streamparse_run_path
        # Update python paths in spouts
        for thrift_spout in topology_class.thrift_spouts.values():
            inner_shell = thrift_spout.spout_object.shell
            if isinstance(inner_shell, ShellComponent):
                if "streamparse_run" in inner_shell.execution_command:
                    inner_shell.execution_command = streamparse_run_path

    # In case we're overriding things, let's save the original name
    options["topology.original_name"] = name

    # Set parallelism based on env_name if necessary
    for thrift_component in chain(
        topology_class.thrift_bolts.values(), topology_class.thrift_spouts.values()
    ):
        par_hint = thrift_component.common.parallelism_hint
        if isinstance(par_hint, dict):
            thrift_component.common.parallelism_hint = par_hint.get(env_name)

    if local_jar_path:
        print(f"Using prebuilt JAR: {local_jar_path}")
    elif not remote_jar_path:
        # Check topology for JVM stuff to see if we need to create uber-jar
        if simple_jar:
            simple_jar = not any(
                isinstance(spec, JavaComponentSpec) for spec in topology_class.specs
            )

        # Prepare a JAR that doesn't have Storm dependencies packaged
        local_jar_path = jar_for_deploy(simple_jar=simple_jar)

    if name != override_name:
        print(f'Deploying "{name}" topology with name "{override_name}"...')
    else:
        print(f'Deploying "{name}" topology...')
    sys.stdout.flush()
    # Use ssh tunnel with Nimbus if use_ssh_for_nimbus is unspecified or True
    with ssh_tunnel(env_config) as (host, port):
        nimbus_client = get_nimbus_client(
            env_config, host=host, port=port, timeout=timeout
        )
        if remote_jar_path:
            print(f"Reusing remote JAR on Nimbus server at path: {remote_jar_path}")
        else:
            remote_jar_path = _upload_jar(nimbus_client, local_jar_path)
        _kill_existing_topology(override_name, force, wait, nimbus_client)
        _submit_topology(
            override_name,
            topology_class,
            remote_jar_path,
            config,
            env_config,
            nimbus_client,
            options=options,
            active=active,
        )
    _post_submit_hooks(override_name, env_name, env_config, options)


def subparser_hook(subparsers):
    """Hook to add subparser for this command."""
    subparser = subparsers.add_parser("submit", description=__doc__, help=main.__doc__)
    subparser.set_defaults(func=main)
    add_ackers(subparser)
    add_config(subparser)
    add_debug(subparser)
    add_environment(subparser)
    subparser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force a topology to submit by killing any "
        "currently running topologies with the same "
        "name.",
    )
    subparser.add_argument(
        "-i",
        "--inactive",
        help="Submit topology as inactive instead of active."
        " This is useful if you are migrating the "
        "topology to a new environment and already "
        "have it running actively in an older one.",
        action="store_false",
        dest="active",
    )
    subparser.add_argument(
        "-j",
        "--local_jar_path",
        help="Path to a prebuilt JAR to upload to Nimbus. "
        "This is useful when you have multiple "
        "topologies that all run out of the same JAR, "
        "or you have manually created the JAR.",
    )
    add_name(subparser)
    add_options(subparser)
    add_override_name(subparser)
    add_overwrite_virtualenv(subparser)
    add_pool_size(subparser)
    add_requirements(subparser)
    subparser.add_argument(
        "-R",
        "--remote_jar_path",
        help="Path to a prebuilt JAR that already exists on "
        "your Nimbus server. This is useful when you "
        "have multiple topologies that all run out of "
        "the same JAR, and you do not want to upload it"
        " multiple times.",
    )
    add_timeout(subparser)
    subparser.add_argument(
        "-u",
        "--uber_jar",
        help="Build an Uber-JAR even if you have no Java "
        "components in your topology.  Useful if you "
        "are providing your own seriailzer class.",
        dest="simple_jar",
        action="store_false",
    )
    add_user(subparser)
    add_wait(subparser)
    add_workers(subparser)


def main(args):
    """Submit a Storm topology to Nimbus."""
    env.pool_size = args.pool_size
    submit_topology(
        name=args.name,
        env_name=args.environment,
        options=args.options,
        force=args.force,
        wait=args.wait,
        simple_jar=args.simple_jar,
        override_name=args.override_name,
        requirements_paths=args.requirements,
        local_jar_path=args.local_jar_path,
        remote_jar_path=args.remote_jar_path,
        timeout=args.timeout,
        config_file=args.config,
        overwrite_virtualenv=args.overwrite_virtualenv,
        active=args.active,
    )



import argparse
import unittest

from streamparse.cli.worker_uptime import subparser_hook


def test_subparser_hook():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    subparser_hook(subparsers)

    subcommands = parser._optionals._actions[1].choices.keys()
    assert "worker_uptime" in subcommands



"""
Module to add streamparse-specific extensions to pystorm Bolt classes
"""

import pystorm

from ..dsl.bolt import JavaBoltSpec, ShellBoltSpec
from .component import Component


class JavaBolt(Component):
    @classmethod
    def spec(
        cls,
        name=None,
        serialized_java=None,
        full_class_name=None,
        args_list=None,
        inputs=None,
        par=1,
        config=None,
        outputs=None,
    ):
        """Create a :class:`JavaBoltSpec` for a Java Bolt.

        This spec represents this Bolt in a :class:`~streamparse.Topology`.

        You must add the appropriate entries to your classpath by editing your
        project's ``project.clj`` file in order for this to work.

        :param name:   Name of this Bolt.  Defaults to name of
                       :class:`~streamparse.Topology` attribute this is assigned
                       to.
        :type name:    `str`
        :param serialized_java:  Serialized Java code representing the class.
                                 You must either specify this, or
                                 both ``full_class_name`` and ``args_list``.
        :type serialized_java: `bytes`
        :param full_class_name: Fully qualified class name (including the
                                package name)
        :type full_class_name: `str`
        :param args_list: A list of arguments to be passed to the constructor of
                          this class.
        :type args_list: `list` of basic data types
        :param inputs: Streams that feed into this Bolt.

                       Two forms of this are acceptable:

                       1.  A `dict` mapping from
                           :class:`~streamparse.dsl.component.ComponentSpec` to
                           :class:`~streamparse.Grouping`.
                       2.  A `list` of :class:`~streamparse.Stream` or
                           :class:`~streamparse.dsl.component.ComponentSpec`.
        :param par:    Parallelism hint for this Bolt.  For Python
                       Components, this works out to be the number of Python
                       processes running it in the the topology (across all
                       machines).  See :ref:`parallelism`.
        :type par:     `int`
        :param config: Component-specific config settings to pass to Storm.
        :type config:  `dict`
        :param outputs: Outputs this JavaBolt will produce. Acceptable forms
                        are:

                        1.  A `list` of :class:`~streamparse.Stream` objects
                            describing the fields output on each stream.
                        2.  A `list` of `str` representing the fields output on
                            the ``default`` stream.
        """
        return JavaBoltSpec(
            cls,
            name=name,
            serialized_java=serialized_java,
            full_class_name=full_class_name,
            args_list=args_list,
            inputs=inputs,
            par=par,
            config=config,
            outputs=outputs,
        )


class ShellBolt(Component):
    """A Bolt that is started by running a command with a script argument."""

    @classmethod
    def spec(
        cls,
        name=None,
        command=None,
        script=None,
        inputs=None,
        par=None,
        config=None,
        outputs=None,
    ):
        """Create a :class:`ShellBoltSpec` for a non-Java, non-Python Bolt.

        If you want to create a spec for a Python Bolt, use
        :meth:`~streamparse.dsl.bolt.Bolt.spec`.

        This spec represents this Bolt in a :class:`~streamparse.Topology`.

        :param name:   Name of this Bolt.  Defaults to name of
                       :class:`~streamparse.Topology` attribute this is assigned
                       to.
        :type name:    `str`
        :param command:  Path to command the Storm will execute.
        :type command: `str`
        :param script: Arguments to `command`.  Multiple arguments should just
                       be separated by spaces.
        :type script: `str`
        :param inputs: Streams that feed into this Bolt.

                       Two forms of this are acceptable:

                       1.  A `dict` mapping from
                           :class:`~streamparse.dsl.component.ComponentSpec` to
                           :class:`~streamparse.Grouping`.
                       2.  A `list` of :class:`~streamparse.Stream` or
                           :class:`~streamparse.dsl.component.ComponentSpec`.
        :param par:    Parallelism hint for this Bolt.  For shell
                       Components, this works out to be the number of running it
                       in the the topology (across all machines).
                       See :ref:`parallelism`.
        :type par:     `int`
        :param config: Component-specific config settings to pass to Storm.
        :type config:  `dict`
        :param outputs: Outputs this ShellBolt will produce. Acceptable forms
                        are:

                        1.  A `list` of :class:`~streamparse.Stream` objects
                            describing the fields output on each stream.
                        2.  A `list` of `str` representing the fields output on
                            the ``default`` stream.
        """
        return ShellBoltSpec(
            cls,
            command=command,
            script=script,
            name=name,
            inputs=inputs,
            par=par,
            config=config,
            outputs=outputs,
        )


class Bolt(pystorm.bolt.Bolt, ShellBolt):
    """pystorm Bolt with streamparse-specific additions"""

    @classmethod
    def spec(cls, name=None, inputs=None, par=None, config=None):
        """Create a :class:`~ShellBoltSpec` for a Python Bolt.

        This spec represents this Bolt in a :class:`~streamparse.Topology`.

        :param name:   Name of this Bolt.  Defaults to name of
                       :class:`~streamparse.Topology` attribute this is assigned
                       to.
        :type name:    `str`
        :param inputs: Streams that feed into this Bolt.

                       Two forms of this are acceptable:

                       1.  A `dict` mapping from
                           :class:`~streamparse.dsl.component.ComponentSpec` to
                           :class:`~streamparse.Grouping`.
                       2.  A `list` of :class:`~streamparse.Stream` or
                           :class:`~streamparse.dsl.component.ComponentSpec`.
        :param par:    Parallelism hint for this Bolt.  For Python
                       Components, this works out to be the number of Python
                       processes running it in the the topology (across all
                       machines).  See :ref:`parallelism`.

                       .. note::
                           This can also be specified as an attribute of your
                           :class:`~Bolt` subclass.

        :type par:     `int`
        :param config: Component-specific config settings to pass to Storm.

                       .. note::
                           This can also be specified as an attribute of your
                           :class:`~Bolt` subclass.

        :type config:  `dict`

        .. note::
            This method does not take a ``outputs`` argument because
            ``outputs`` should be an attribute of your :class:`~Bolt` subclass.
        """
        return ShellBoltSpec(
            cls,
            command="streamparse_run",
            script=f"{cls.__module__}.{cls.__name__}",
            name=name,
            inputs=inputs,
            par=par,
            config=config,
            outputs=cls.outputs,
        )


class BatchingBolt(pystorm.bolt.BatchingBolt, Bolt):
    """pystorm BatchingBolt with streamparse-specific additions"""

    pass


class TicklessBatchingBolt(pystorm.bolt.TicklessBatchingBolt, BatchingBolt):
    """pystorm TicklessBatchingBolt with streamparse-specific additions"""

    pass



"""
Helper script for starting up bolts and spouts.
"""

import argparse
import importlib
import os
import sys

from pystorm.component import _SERIALIZERS

RESOURCES_PATH = "resources"


def main():
    """main entry point for Python bolts and spouts"""
    parser = argparse.ArgumentParser(
        description="Run a bolt/spout class",
        epilog="This is internal to streamparse "
        "and is used to run spout and bolt "
        "classes via ``python -m "
        "streamparse.run <class name>``.",
    )
    parser.add_argument("target_class", help="The bolt/spout class to start.")
    parser.add_argument(
        "-s",
        "--serializer",
        help="The serialization protocol to use to talk to " "Storm.",
        choices=_SERIALIZERS.keys(),
        default="json",
    )
    # Storm sends everything as one string, which is not great
    if len(sys.argv) == 2:
        sys.argv = [sys.argv[0]] + sys.argv[1].split()
    args = parser.parse_args()
    mod_name, cls_name = args.target_class.rsplit(".", 1)
    # Add current directory to sys.path so imports will work
    import_path = os.getcwd()  # Storm <= 1.0.2
    if RESOURCES_PATH in next(os.walk(import_path))[1] and os.path.isfile(
        os.path.join(
            import_path, RESOURCES_PATH, mod_name.replace(".", os.path.sep) + ".py"
        )
    ):
        import_path = os.path.join(import_path, RESOURCES_PATH)  # Storm >= 1.0.3
    sys.path.append(import_path)
    # Import module
    mod = importlib.import_module(mod_name)
    # Get class from module and run it
    cls = getattr(mod, cls_name)
    cls(serializer=args.serializer).run()


if __name__ == "__main__":
    main()



"""
Kill the specified Storm topology.
"""

from ..thrift import KillOptions
from ..util import (
    get_topology_definition,
    get_env_config,
    get_nimbus_client,
    ssh_tunnel,
)
from .common import add_config, add_environment, add_name, add_timeout, add_wait


def _kill_topology(topology_name, nimbus_client, wait=None):
    kill_opts = KillOptions(wait_secs=wait)
    nimbus_client.killTopologyWithOpts(name=topology_name, options=kill_opts)


def kill_topology(
    topology_name=None, env_name=None, wait=None, timeout=None, config_file=None
):
    # For kill, we allow any topology name to be specified, because people
    # should be able to kill topologies not in their local branch
    if topology_name is None:
        topology_name = get_topology_definition(topology_name, config_file=config_file)[
            0
        ]
    env_name, env_config = get_env_config(env_name, config_file=config_file)
    # Use ssh tunnel with Nimbus if use_ssh_for_nimbus is unspecified or True
    with ssh_tunnel(env_config) as (host, port):
        nimbus_client = get_nimbus_client(
            env_config, host=host, port=port, timeout=timeout
        )
        return _kill_topology(topology_name, nimbus_client, wait=wait)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser("kill", description=__doc__, help=main.__doc__)
    subparser.set_defaults(func=main)
    add_config(subparser)
    add_environment(subparser)
    add_name(subparser)
    add_timeout(subparser)
    add_wait(subparser)


def main(args):
    """ Kill the specified Storm topology """
    kill_topology(
        topology_name=args.name,
        env_name=args.environment,
        wait=args.wait,
        timeout=args.timeout,
        config_file=args.config,
    )



"""
Functions for adding common CLI arguments to argparse sub-commands.
"""
import argparse
import copy
import warnings

from ruamel import yaml

from streamparse.util import get_storm_workers


class _StoreDictAction(argparse.Action):
    """Action for storing key=val option strings as a single dict."""

    def __init__(
        self,
        option_strings,
        dest,
        nargs=None,
        const=None,
        default=None,
        type=None,
        choices=None,
        required=False,
        help=None,
        metavar=None,
    ):
        if nargs == 0:
            raise ValueError("nargs for store_dict actions must be > 0")
        if const is not None and nargs != "?":
            raise ValueError('nargs must be "?" to supply const')
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            const=const,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        if getattr(namespace, self.dest, None) is None:
            setattr(namespace, self.dest, {})
        # Only doing a copy here because that's what _AppendAction does
        items = copy.copy(getattr(namespace, self.dest))
        key, val = values.split("=", 1)
        if yaml.version_info < (0, 15):
            items[key] = yaml.safe_load(val)
        else:
            yml = yaml.YAML(typ="safe", pure=True)
            items[key] = yml.load(val)
        setattr(namespace, self.dest, items)


def option_alias(option):
    """Returns a function that will create option=val for _StoreDictAction."""

    def _create_key_val_str(val):
        return f"{option}={val}"

    return _create_key_val_str


def add_ackers(parser):
    """ Add --ackers option to parser """
    parser.add_argument(
        "-a",
        "--ackers",
        help="Set number of acker executors for your topology. "
        "Defaults to the number of worker nodes in your "
        "Storm environment.",
        type=option_alias("topology.acker.executors"),
        action=_StoreDictAction,
        dest="options",
    )


def add_config(parser):
    """ Add --config option to parser """
    parser.add_argument(
        "--config", help="Specify path to config.json", type=argparse.FileType("r")
    )


def add_debug(parser):
    """ Add --debug option to parser """
    parser.add_argument(
        "-d",
        "--debug",
        help="Set topology.debug and produce debugging output.",
        type=option_alias("topology.debug"),
        action=_StoreDictAction,
        dest="options",
        const="true",
        nargs="?",
    )


def add_environment(parser):
    """ Add --environment option to parser """
    parser.add_argument(
        "-e",
        "--environment",
        help="The environment to use for the command.  "
        "Corresponds to an environment in your "
        '"envs" dictionary in config.json.  If you '
        "only have one environment specified, "
        "streamparse will automatically use this.",
    )


def add_name(parser):
    """ Add --name option to parser """
    parser.add_argument(
        "-n",
        "--name",
        help="The name of the topology to act on.  If you have "
        'only one topology defined in your "topologies" '
        "directory, streamparse will use it "
        "automatically.",
    )


def add_options(parser):
    """ Add --option options to parser """
    parser.add_argument(
        "-o",
        "--option",
        dest="options",
        action=_StoreDictAction,
        help="Topology option to pass on to Storm. For example,"
        ' "-o topology.debug=true" is equivalent to '
        '"--debug".  May be repeated multiple for multiple'
        " options.",
    )


def add_override_name(parser):
    """ Add --override_name option to parser """
    parser.add_argument(
        "-N",
        "--override_name",
        help="For operations such as creating virtualenvs and "
        "killing/submitting topologies, use this value "
        "instead of NAME.  This is useful if you want to "
        "submit the same topology twice without having to "
        "duplicate the topology file.",
    )


def add_overwrite_virtualenv(parser):
    """ Add --overwrite_virtualenv option to parser """
    parser.add_argument(
        "--overwrite_virtualenv",
        help="Create the virtualenv even if it already exists."
        " This is useful when you have changed your "
        "virtualenv_flags.",
        action="store_true",
    )


def add_pattern(parser):
    """ Add --pattern option to parser """
    parser.add_argument("--pattern", help="Pattern of log files to operate on.")


def add_pool_size(parser):
    """ Add --pool_size option to parser """
    parser.add_argument(
        "--pool_size",
        help="Number of simultaneous SSH connections to use when updating "
        "virtualenvs, removing logs, or tailing logs.",
        default=10,
        type=int,
    )


def add_requirements(parser):
    """ Add --requirements option to parser """
    parser.add_argument(
        "-r",
        "--requirements",
        nargs="*",
        help="Path to pip-style requirements file specifying "
        "the dependencies to use for creating the "
        "virtualenv for this topology.  If unspecified, "
        "streamparse will look for a file called NAME.txt "
        "in the directory specified by the "
        "virtualenv_specs setting in config.json.",
    )


def add_simple_jar(parser):
    """ Add --simple_jar option to parser. """
    parser.add_argument(
        "-s",
        "--simple_jar",
        action="store_true",
        help="Instead of creating an Uber-JAR for the "
        "topology, which contains all of its JVM "
        "dependencies, create a simple JAR with just the "
        "code for the project.  This is useful when your "
        "project is pure Python and has no JVM "
        "dependencies.",
    )


def add_timeout(parser):
    """ Add --timeout option to parser """
    parser.add_argument(
        "--timeout",
        type=int,
        default=7000,
        help="Milliseconds to wait for Nimbus to respond. " "(default: %(default)s)",
    )


def add_user(parser, allow_short=False):
    """Add --user option to parser

    Set allow_short to add -u as well.
    """
    args = ["--user"]
    if allow_short:
        args.insert(0, "-u")

    kwargs = {
        "help": "User argument to sudo when creating and deleting virtualenvs.",
        "default": None,
        "type": option_alias("sudo_user"),
        "dest": "options",
        "action": _StoreDictAction,
    }

    parser.add_argument(*args, **kwargs)


def add_wait(parser):
    """ Add --wait option to parser """
    parser.add_argument(
        "--wait",
        type=int,
        default=5,
        help="Seconds to wait before killing topology. " "(default: %(default)s)",
    )


def add_workers(parser):
    """ Add --workers option to parser """
    parser.add_argument(
        "-w",
        "--workers",
        help="Set number of Storm workers for your topology. "
        "Defaults to the number of worker nodes in your "
        "Storm environment.",
        type=option_alias("topology.workers"),
        action=_StoreDictAction,
        dest="options",
    )


VIRTUALENV_OPTIONS = (
    "install_virtualenv",
    "use_virtualenv",
    "virtualenv_flags",
    "virtualenv_root",
    "virtualenv_name",
)


def resolve_options(
    cli_options, env_config, topology_class, topology_name, local_only=False
):
    """Resolve potentially conflicting Storm options from three sources:

    CLI options > Topology options > config.json options

    :param local_only: Whether or not we should talk to Nimbus to get Storm
                       workers and other info.
    """
    storm_options = {}

    # Start with environment options
    storm_options.update(env_config.get("options", {}))

    # Set topology.python.path
    if env_config.get("use_virtualenv", True):
        python_path = "/".join(
            [env_config["virtualenv_root"], topology_name, "bin", "python"]
        )
        # This setting is for information purposes only, and is not actually
        # read by any pystorm or streamparse code.
        storm_options["topology.python.path"] = python_path

    # Set logging options based on environment config
    log_config = env_config.get("log", {})
    log_path = log_config.get("path") or env_config.get("log_path")
    log_file = log_config.get("file") or env_config.get("log_file")
    if log_path:
        storm_options["pystorm.log.path"] = log_path
    if log_file:
        storm_options["pystorm.log.file"] = log_file
    if isinstance(log_config.get("max_bytes"), int):
        storm_options["pystorm.log.max_bytes"] = log_config["max_bytes"]
    if isinstance(log_config.get("backup_count"), int):
        storm_options["pystorm.log.backup_count"] = log_config["backup_count"]
    if isinstance(log_config.get("level"), str):
        storm_options["pystorm.log.level"] = log_config["level"].lower()

    # Make sure virtualenv options are present here
    for venv_option in VIRTUALENV_OPTIONS:
        if venv_option in env_config:
            storm_options[venv_option] = env_config[venv_option]

    # Set sudo_user default to SSH user if we have one
    storm_options["sudo_user"] = env_config.get("user", None)

    # Override options with topology options
    storm_options.update(topology_class.config)

    # Override options with CLI options
    storm_options.update(cli_options or {})

    # Set log level to debug if topology.debug is set
    if storm_options.get("topology.debug", False):
        storm_options["pystorm.log.level"] = "debug"

    # If ackers and executors still aren't set, use number of worker nodes
    if not local_only:
        if not storm_options.get("storm.workers.list"):
            storm_options["storm.workers.list"] = get_storm_workers(env_config)
        elif isinstance(storm_options["storm.workers.list"], str):
            storm_options["storm.workers.list"] = storm_options[
                "storm.workers.list"
            ].split(",")
        num_storm_workers = len(storm_options["storm.workers.list"])
    else:
        storm_options["storm.workers.list"] = []
        num_storm_workers = 1
    if storm_options.get("topology.acker.executors") is None:
        storm_options["topology.acker.executors"] = num_storm_workers
    if storm_options.get("topology.workers") is None:
        storm_options["topology.workers"] = num_storm_workers

    # If sudo_user was not present anywhere, set it to "root"
    storm_options.setdefault("sudo_user", "root")

    return storm_options


def warn_about_deprecated_user(user, func_name):
    if user is not None:
        warnings.warn(
            (
                "The 'user' argument to '{}' will be removed in the next "
                "major release of streamparse. Provide the 'sudo_user' key to"
                " the 'options' dict argument instead."
            ).format(func_name),
            DeprecationWarning,
        )
